"""
From the spec:

    The eight [sic! It's nine!] primary public blocks are TAXA, CHARACTERS, UNALIGNED, DISTANCES,
    SETS, ASSUMPTIONS, CODONS, TREES, and NOTES.
"""
from .base import Block  # noqa: F401
from .assumptions import Assumptions  # noqa: F401
from .characters import Characters, Data  # noqa: F401
from .codons import Codons  # noqa: F401
from .distances import Distances  # noqa: F401
from .notes import Notes  # noqa: F401
from .sets import Sets  # noqa: F401
from .taxa import Taxa  # noqa: F401
from .trees import Trees  # noqa: F401
from .unaligned import Unaligned  # noqa: F401



import functools
import itertools

from .tokenizer import TokenType, iter_tokens, Token, get_name


class Command(tuple):
    """
    From the specification:

        A command is a collection of tokens terminated by a semicolon. Commands cannot contain
        semicolons, except as terminators, unless the semicolons are contained within a comment
        or within a quoted token consisting of more than one text character.
    """
    @functools.cached_property
    def name(self):
        return get_name(self)

    def __str__(self):
        return ''.join(str(t) for t in self)

    def __eq__(self, other):  # To make command removal in Nexus.replace_block work.
        return id(self) == id(other)

    @classmethod
    def from_name_and_payload(cls, name, payload='', comment=None, in_block=False):
        tokens = []
        if comment:
            tokens.extend([Token('\n', TokenType.WHITESPACE), Token(comment, TokenType.COMMENT)])
        tokens.append(Token('\n', TokenType.WHITESPACE))
        name = list(iter_tokens(iter(name)))
        assert len(name) == 1 and name[0].type == TokenType.WORD
        tokens.extend(name)
        semicolons = 0
        if payload:
            tokens.append(Token(' ', TokenType.WHITESPACE))
            payload = list(iter_tokens(iter(payload)))
            semicolons = len([t for t in payload if t.is_semicolon])
            assert semicolons == 0 or (semicolons == 1 and payload[-1].is_semicolon)
            tokens.extend(payload)
        if semicolons == 0:
            tokens.append(Token(';', TokenType.PUNCTUATION))
        return cls(tokens)

    def without_comments(self) -> 'Command':
        """
        Returns a `Command` instance based on the tokens of `self`, but with all non-command
        comments removed.

        .. note::

            Command comment.—A comment containing a command is a command comment. Comments
            containing commands are indicated by a special symbol (\\ or &) as the first text
            character within the comment.

        .. warning::

            This will also remove comments in Newick representations of trees if these do not
            follow the NEXUS "command comment" conventions.
        """
        return Command(t for t in self if t.type != TokenType.COMMENT or t.text[0] in '&\\')

    def with_normalised_whitespace(self):
        comments, name, postcomments = [], [], []
        i, j = -1, -1
        for i, token in enumerate(self):
            if token.type == TokenType.WORD:
                name.append(token)
                break
            if token.type == TokenType.COMMENT:
                comments.append(token)
        for j, token in enumerate(self):
            if j <= i:
                continue
            if token.type == TokenType.WORD:
                name.append(token)
            elif token.type == TokenType.COMMENT:
                postcomments.append(token)
            else:
                break

        tokens = []
        if comments:
            tokens.append(Token('\n', TokenType.WHITESPACE))
            tokens.extend(comments)
        tokens.append(Token(
            '\n' if self.is_beginblock or self.is_endblock else '\n\t', TokenType.WHITESPACE))
        tokens.append((Token(''.join(t.text.upper() for t in name), TokenType.WORD)))
        if postcomments:
            tokens.extend(postcomments)
        tokens.extend(self[j:])
        return Command(tokens)

    @property
    def is_beginblock(self) -> bool:
        return self.name == 'BEGIN'

    @property
    def is_endblock(self) -> bool:
        # In MacClade, PAUP, and COMPONENT, the ENDBLOCK command has been used as
        # a synonym of the END command.
        return self.name in ['END', 'ENDBLOCK']

    def iter_payload_tokens(self, type=None):
        found = False
        for t in itertools.dropwhile(
                lambda t: t.type != TokenType.WHITESPACE,
                itertools.dropwhile(
                    lambda t: t.type in {TokenType.WHITESPACE, TokenType.COMMENT}, self[:-1])):
            if not found and t.type == TokenType.WHITESPACE:
                # Drop the first whitespace token, too.
                continue
            found = True
            if type is None or (type == t.type):
                yield t



"""

Punctuation
~~~~~~~~~~~

    Any of the following text characters are considered punctuation at some times:

    ..code-block::

        ( ) [ ] { } / \\ , ; : = * ' " " + - < >

    The following punctuation marks have special properties: [ ] do not break a
    word; + and - are allowed as state symbols, but none of the rest are allowed; - is
    considered punctuation except were it is the minus sign in a negative number.
"""
import enum
import itertools
import dataclasses

__all__ = [
    'TokenType', 'Token', 'iter_tokens', 'get_name', 'iter_words_and_punctuation', 'Word',
    'iter_delimited', 'iter_lines', 'word_after_equals', 'iter_key_value_pairs']

import typing

QUOTE = "'"
COMMENT = {'[': 1, ']': -1}
PUNCTUATION = r'(){}/\,;:=*"+-<>'
WHITESPACE = '\t\r\n '
BOOLEAN = {}
for s in 'yes 1 true'.split():
    BOOLEAN[s] = True
for s in 'no 0 false'.split():
    BOOLEAN[s] = False


class TokenType(enum.Enum):
    WORD = 1
    QWORD = 2
    COMMENT = 3
    PUNCTUATION = 4
    WHITESPACE = 5


@dataclasses.dataclass
class Token:
    """
    We parse Nexus in one pass, storing the data as list of tokens with enough
    information to extract relevant parts from this list later on.
    """
    __slots__ = ['text', 'type']
    text: str
    type: TokenType

    @classmethod
    def from_text(cls, tokens, type=None):
        return cls(
            text=''.join(tokens),
            type=type or (TokenType.WHITESPACE if tokens[-1] in WHITESPACE else TokenType.WORD))

    @property
    def is_semicolon(self):
        return self.type == TokenType.PUNCTUATION and self.text == ';'

    @property
    def is_whitespace(self):
        return self.type == TokenType.WHITESPACE

    @property
    def is_newline(self):
        return self.is_whitespace and any(c in self.text for c in '\r\n')

    @property
    def is_punctuation(self):
        return self.type == TokenType.PUNCTUATION

    def __str__(self):
        if self.type == TokenType.COMMENT:
            return '[{}]'.format(self.text)
        if self.type == TokenType.QWORD:
            return "{}{}{}".format(QUOTE, self.text.replace(QUOTE, QUOTE + QUOTE), QUOTE)
        return self.text


def iter_tokens(s):
    token, lookahead = [], None

    while True:
        try:
            c = lookahead or next(s)
            lookahead = None

            if c == QUOTE:  # A quoted string.
                doublequote = False
                if token:
                    if token[-1] in WHITESPACE:
                        yield Token.from_text(token, TokenType.WHITESPACE)
                    else:
                        raise ValueError()  # pragma: no cover
                    token = []

                while 1:
                    c = lookahead or next(s)
                    lookahead = None
                    while c != QUOTE:
                        token.append(c)
                        c = next(s)

                    if doublequote:
                        token.append(c)
                        doublequote = False
                    else:
                        try:
                            lookahead = next(s)
                        except StopIteration:
                            lookahead = None
                        if lookahead == QUOTE:
                            doublequote = True
                        else:  # End of quoted string
                            yield Token.from_text(token, type=TokenType.QWORD)
                            token = []
                            break
                continue

            if c == '[':  # A comment.
                if token:
                    yield Token.from_text(token)
                    token = []
                commentlevel = 1
                while 1:
                    c = next(s)
                    while c not in COMMENT:
                        token.append(c)
                        c = next(s)
                    commentlevel += COMMENT[c]
                    if commentlevel == 0:
                        yield Token.from_text(token, TokenType.COMMENT)
                        token = []
                        break
                    else:
                        token.append(c)
                continue

            if c in WHITESPACE:
                if token and (token[-1] not in WHITESPACE):
                    yield Token.from_text(token, TokenType.WORD)
                    token = []
                token.append(c)
                continue

            if c in PUNCTUATION:
                if token:
                    yield Token.from_text(token)
                    token = []
                yield Token.from_text(c, TokenType.PUNCTUATION)
                continue

            if token and (token[-1] in WHITESPACE):
                yield Token.from_text(token, TokenType.WHITESPACE)
                token = []
            token.append(c)

        except StopIteration:
            break

    if token:
        yield Token.from_text(token)


def get_name(tokens):
    """
    The Nexus spec allows comments **in** block or command names. This function takes this into
    account when assembling a name from an iterable of tokens.
    """
    res = ''
    for t in itertools.dropwhile(lambda t: t.type != TokenType.WORD, tokens):
        if res:
            # We already encountered one word.
            if t.type == TokenType.WORD:
                res += t.text
            elif t.type != TokenType.COMMENT:
                break
        else:
            res += t.text
    return res.upper()


class Word(str):
    def __eq__(self, other):
        return self.replace(' ', '_') == other.replace(' ', '_') \
            if isinstance(other, str) else False

    def as_nexus_string(self):
        must_quote = False
        for chars in [WHITESPACE, COMMENT, PUNCTUATION, QUOTE]:
            if any(c in self for c in chars):
                must_quote = True
                break
        if must_quote:
            return "{}{}{}".format(QUOTE, self.replace(QUOTE, QUOTE + QUOTE), QUOTE)
        return self

    def __hash__(self):
        return hash(str(self))


def iter_words_and_punctuation(tokens, allow_punctuation_in_word=None, nexus=None):
    """
    Word

    From the spec:

        Except for special cases involving quotes or comments, a NEXUS word is any string of text
        characters that is bounded by whitespace or punctuation and that does not contain
        whitespace or punctuation. If the first character of a word is a single quote, then
        the word ends with the next single quote (unless that single quote is in a pair of
        consecutive single quotes; if so, then the word ends at the first unpaired single quote).
        Any character, including punctuation and whitespace, may be contained within a quoted word.
        A word cannot consist of only whitespace and punctuation. On each of the following lines
        is a single legal word:

        .. code-block::

            Bembidion
            B._zephyrum
            'John''s sparrow (eastern) '

        Underscores are considered equivalent to blank spaces, except that underscores are dark
        characters and blank spaces are whitespace. Thus, a program encountering "B._zephyrum" and
        "'B. zephyrum'" should judge them to be identical.  Any doubled single quotes within a
        quoted word should be converted to single quotes within the program. Thus, 'John''s' would
        be treated internally within a program as John's. Of course, if the program writes out
        such a word to a NEXUS file, the quote should be redoubled, and quotes should be written
        around the word: 'John' 's.
        A string of dark characters is broken into several words by whitespace and any punctuation
        other than brackets, unless the string is surrounded by quotes.
        Thus, "[\\i]Bembidion_velox[\\p]_Linnaeus" consists of one word, as does
        '- - A single word', whereas "two()words" consists of four tokens: the word "two",
        two punctuation characters, and the word "words".
    """
    allow_punctuation_in_word = allow_punctuation_in_word or ''
    if nexus is not None:
        if nexus.cfg.hyphenminus_is_text:
            allow_punctuation_in_word += '-'
        if nexus.cfg.asterisk_is_text:
            allow_punctuation_in_word += "*"
    word = ''
    for i, token in enumerate(tokens):
        if token.type == TokenType.QWORD:
            assert not word
            yield token.text
        elif token.type == TokenType.WORD:
            word += token.text
        elif token.is_whitespace:
            if word:
                yield Word(word)
                word = ''
        elif token.is_punctuation:
            if token.text in allow_punctuation_in_word:
                word += token.text
            else:
                if word:
                    yield Word(word)
                    word = ''
                yield token
    if word:
        yield Word(word)


def iter_key_value_pairs(tokens, allow_punctuation_in_word=None):
    """
    :param tokens:
    :param allow_punctuation_in_word:
    :return:
    """
    key, e, value, b = None, False, [], False
    #
    # FIXME: need to pass nexus to iter_words_and_punctuation
    #
    for t in iter_words_and_punctuation(
            tokens, allow_punctuation_in_word=allow_punctuation_in_word):
        if key is None:
            assert isinstance(t, str)
            key = t
        elif not e:
            assert isinstance(t, Token) and t.text == '='
            e = True
        else:
            if isinstance(t, Token):
                if t.text == '(':
                    assert not value
                    b = True
                elif t.text == ')':
                    assert b
                    yield key, value
                    key, e, value, b = None, False, [], False
                else:
                    assert b
                    value.append(t.text)
            else:
                if b:
                    value.append(t)
                else:
                    assert not value
                    yield key, [t]
                    key, e, value, b = None, False, [], False


def word_after_equals(words_and_punctuation: typing.Generator) -> str:
    n = next(words_and_punctuation)
    assert n.text == '='
    res = next(words_and_punctuation)
    return res if isinstance(res, str) else res.text


def iter_lines(tokens):
    line = []
    for t in tokens:
        if t.is_newline:
            if not all(tt.type in (TokenType.WHITESPACE, TokenType.COMMENT) for tt in line):
                yield line
            line = []
        else:
            line.append(t)
    if line:
        if not all(tt.type in (TokenType.WHITESPACE, TokenType.COMMENT) for tt in line):
            yield line


def iter_delimited(start, words_and_punctuation, delimiter='"', allow_single_word=False):
    startchar = delimiter[0]
    endchar = startchar if len(delimiter) == 1 else delimiter[1]

    if isinstance(start, str) and start != startchar:
        if allow_single_word:
            yield start
            return
        raise ValueError("No delimiter found!")  # pragma: no cover
    else:
        start = start or next(words_and_punctuation)
        assert getattr(start, 'text', start) == startchar
        w = next(words_and_punctuation)
        while not (isinstance(w, Token) and w.text == endchar):
            yield w
            w = next(words_and_punctuation)



import typing
import collections

from commonnexus.tokenizer import iter_words_and_punctuation, Word
from .base import Block, Payload


class Dimensions(Payload):
    """
    The NTAX subcommand of the DIMENSIONS command indicates the number of taxa. The NEXUS standard
    does not impose limits on number of taxa; a limit may be imposed by particular computer
    programs.

    :ivar int ntax: The number of taxa.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        text = [t.text for t in tokens if not t.is_whitespace]
        assert text[0].upper() == 'NTAX' and text[1] == '='
        self.ntax = int(text[2])


class Taxlabels(Payload):
    """
    This command defines taxa, specifies their names, and determines their order:

    .. code-block::

        TAXLABELS Fungus Insect Mammal;

    Taxon names are single NEXUS words. They must not correspond to another taxon name or number;
    thus, 1 is not a valid name for the second taxon listed. The standard defines no limit to
    their length, although individual programs might impose restrictions.

    Taxa may also be defined in the CHARACTERS, UNALIGNED, and DISTANCES blocks if the NEWTAXA
    token is included in the DIMENSIONS command; see the descriptions of those blocks for details.

    :ivar typing.Dict[int, str] labels: Mapping of taxon number to taxon label.

    The taxon number is the number of a taxon, as defined by its position in a TAXLABELS
    command. [...] For example, the third taxon listed in TAXLABELS is taxon number 3.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.labels = collections.OrderedDict(
            [(n, w) for n, w in enumerate(
                iter_words_and_punctuation(tokens, nexus=nexus), start=1)])
        assert len(self.labels) == len(set(self.labels.values())), 'Duplicates in TAXLABELS'
        assert not set(str(n) for n in self.labels).intersection(self.labels.values()), \
            'Numbers as labels'


class Taxa(Block):
    """
    The :class:`TAXA block <Taxa>` specifies information about taxa.

    .. code-block::

        BEGIN TAXA;
            DIMENSIONS NTAX=number-of-taxa;
            TAXLABELS taxon-name [taxon-name ...];
        END;

    :class:`DIMENSIONS <Dimensions>` must appear before :class:`TAXLABELS <Taxlabels>`.
    Only one of each command is allowed per block.
    """
    __commands__ = [Dimensions, Taxlabels]

    @classmethod
    def from_data(cls,
                  labels: typing.Iterable,
                  TITLE: typing.Optional[str] = None,
                  ID: typing.Optional[str] = None,
                  LINK: typing.Optional[typing.Union[str, typing.Tuple[str, str]]] = None,
                  nexus=None) -> 'Block':
        return cls.from_commands([
            ('DIMENSIONS', 'NTAX={}'.format(len(labels))),
            ('TAXLABELS', ' '.join(Word(w).as_nexus_string() for w in labels)),
        ], nexus=nexus, TITLE=TITLE, ID=ID, LINK=LINK)



import typing

from .base import Block, Payload
from commonnexus.tokenizer import iter_key_value_pairs, Word


class Text(Payload):
    """
    This command allows text to be attached to various objects.

    The SOURCE subcommand indicates the location of the text. The INLINE option indicates that the
    text is present at the end of the TEXT command; the FILE option indicates that it is in a
    separate file (the name of which is then specified in the TEXT subcommand); the RESOURCE option
    indicates that it is in the resource fork of the current file, in a resource of type TEXT (the
    numerical ID of which is then specified in the TEXT subcommand).

    For example, in the following

    .. code-block::

        TEXT TAXON=5 CHARACTER=2 TEXT='4 specimens observed';
        TEXT TAXON=Pan TEXT='This genus lives in Africa';
        TEXT CHARACTER=2 TEXT='Perhaps this character should be deleted';
        TEXT CHARACTER=1 STATE=0 TEXT='This state is hard to detect';

    the first command assigns the note "4 specimens observed" to the data entry for taxon 5,
    character 2; the second command assigns the note "Perhaps this character should be deleted" to
    character 2; the third command assigns the note "This genus lives in Africa" to the taxon Pan,
    and the last command assigns the note "This state is hard to detect" to state 0 of character 1.

    The text or source descriptor must be a single NEXUS word. If the text contains NEXUS
    whitespace or punctuation, it needs to be surrounded by single quotes, with any contained
    single quotes converted to a pair of single quotes.

    :ivar typing.List[str] taxons: list of taxon labels or numbers the text relates to.
    """
    def __init__(self, tokens=None, nexus=None, **kw):
        super().__init__(tokens, nexus=nexus)
        self.taxons = None
        self.characters = None
        self.states = None
        self.trees = None
        self.source = None
        self.text = None

        if self._tokens is None:
            for k, v in kw.items():
                setattr(self, k, v)
        else:
            for key, values in iter_key_value_pairs(self._tokens):
                key = key.lower()
                if key in ['taxon', 'character', 'state', 'tree']:
                    key = key + 's'
                else:
                    assert len(values) == 1
                    values = values[0]
                setattr(self, key.lower(), values)
            if self.nexus:
                for key in ['taxon', 'character', 'state', 'tree']:
                    key = key + 's'
                    if getattr(self, key):
                        kw = {}
                        if key == 'states':
                            kw['chars'] = self.characters
                        setattr(self, key, self.nexus.resolve_set_spec(
                            key[:-1].upper(), getattr(self, key), **kw))
            assert self.source in [None, 'FILE', 'INLINE'], self.source

    def as_payload(self):
        res = []
        if self.taxons:
            numbers = ' '.join(
                self.nexus.get_numbers('TAXON', self.taxons) if self.nexus else self.taxons)
            res.append("TAXON={}".format('({})'.format(numbers) if ' ' in numbers else numbers))

        if self.characters:
            numbers = ' '.join(self.nexus.get_numbers('CHARACTER', self.characters)
                               if self.nexus else self.characters)
            res.append("CHARACTER={}".format('({})'.format(numbers) if ' ' in numbers else numbers))

        #
        # FIXME: support STATE!
        #

        if self.trees:
            numbers = ' '.join(
                self.nexus.get_numbers('TREE', self.trees) if self.nexus else self.trees)
            res.append("TREE={}".format('({})'.format(numbers) if ' ' in numbers else numbers))

        res.append('SOURCE=INLINE')
        res.append('TEXT={}'.format(Word(self.text).as_nexus_string()))
        return ' '.join(res)


class Picture(Payload):
    """
    This command allows a picture to be attached to an object.

    The FORMAT subcommand allows specification of the graphics format of the image.
    The SOURCE subcommand indicates the location of the picture. The INLINE option indicates that
    the picture is present at the end of the PICTURE command; the FILE option indicates that it is
    in a separate file (the name of which is then specified in the PICTURE subcommand); the
    RESOURCE option indicates that it is in the resource fork of the current file, in a resource of
    type PICT (the numerical ID of which is then specified in the PICTURE command). The RESOURCE
    option is designed for AppleMacintosh® text files.

    For example, the following command

    .. code-block::

        PICTURE TAXON=5 CHARACTER=2 FORMAT=GIF SouRCE=file PiCTURE=wide.thorax.gif;

    assigns the image in the GIF-formatted file wide.thorax.gif to the data entry for taxon 5,
    character 2.

    The picture or source descriptor must be a single NEXUS word. If the picture contains NEXUS
    whitespace or punctuation, it needs to be surrounded by single quotes, with any contained
    single quotes converted to a pair of single quotes.

    Most graphics formats do not describe pictures using standard text characters. For this reason
    many images cannot be included INLINE in a NEXUS command unless they are converted into text
    characters. The ENCODE subcommand specifies the conversion mechanism used for inline images.

    .. warning::

        Support for encoding of type ``UUENCODE`` will be removed in Python 3.13, because

            `base64 is a modern alternative <https://docs.python.org/3/library/uu.html>`_
    """


class Notes(Block):
    """
    The NOTES block stores notes about various objects in a NEXUS file, including taxa, characters,
    states, and trees:

    .. rst-class:: nexus

        | BEGIN NOTES;
        |   [TEXT
        |     [TAXON=taxon-set]
        |     [CHARACTER=character-set]
        |     [STATE=state-set]
        |     [TREE=tree-set]
        |     SOURCE={INLINE | FILE | RESOURCE} TEXT=text-or-source-descriptor;]
        |   [PICTURE
        |     [TAX0N=taxon-set]
        |     [CHARACTER=character-set]
        |     [STATE= state-set]
        |     [TREE=tree-set]
        |     [FORMAT={PICT | TIFF | EPS | JPEG | GIF}]
        |     [ENCODE={NONE | UUENCODE | BINHEX}]
        |     SOURCE={INLINE | FILE | RESOURCE}
        |     PICTURE=picture-or-source-descriptor; ]
        | END;

    There are no restrictions on the order of commands.

    If the written description of the taxon-set, character-set, state-set, or tree-set contains
    more than one token, it must be enclosed in parentheses, as in the following example:

    .. code-block::

        TEXT TAXON=(1-3) TEXT= 'these taxa from the far north';

    If both a taxon-set and a character-set are specified, then the text or picture applies to
    those characters for those particular taxa. If both a character-set and a state-set are
    specified, then the text or picture applies to those states for those particular characters.

    .. warning::

        PICTURE and ``SOURCE=RESOURCE`` for TEXT is not supported by `commonnexus`.
    """
    __commands__ = {Text, Picture}

    @classmethod
    def from_data(cls,
                  texts: typing.List[typing.Dict[str, typing.Union[typing.List[str], str]]],
                  TITLE: typing.Optional[str] = None,
                  ID: typing.Optional[str] = None,
                  LINK: typing.Optional[typing.Union[str, typing.Tuple[str, str]]] = None,
                  **kw) -> 'Block':
        nexus = kw.pop('nexus', None)
        cmds = [('TEXT', Text(nexus=nexus, **text).as_payload()) for text in texts]
        return cls.from_commands(cmds, nexus=nexus, TITLE=TITLE, ID=ID, LINK=LINK)

    @property
    def texts(self):
        return self.commands['TEXT']

    def get_texts(self, taxon=None, character=None, tree=None):
        res = []
        for text in self.texts:
            if (taxon and text.taxons and taxon in text.taxons) or \
                    (character and text.characters and character in text.characters) or \
                    (tree and text.trees and tree in text.trees):
                res.append(text)
        return res



"""
Basic building blocks of NEXUS files.
"""
import re
import typing
import functools
import collections

from commonnexus.tokenizer import (
    get_name, iter_tokens, iter_words_and_punctuation, word_after_equals, TokenType, Word,
)
from commonnexus.command import Command


class Payload:
    """
    The payload of a Nexus command, i.e. the stuff between command name and final ";".
    """
    __multivalued__ = False

    def __init__(self, tokens, nexus=None):
        self.nexus = nexus
        self._tokens = list(iter_tokens(iter(tokens))) if isinstance(tokens, str) else tokens

    @functools.cached_property
    def comments(self):
        return [t.text for t in self._tokens if t.type == TokenType.COMMENT]

    def format(self, *args, **kw):
        """
        Derived classes may provide functionality to format command data as correct Nexus payload.
        """
        raise NotImplementedError()  # pragma: no cover

    def __str__(self):
        return ''.join(str(t) for t in self._tokens)

    @property
    def lines(self):
        return re.split(r'[\t\r ]*\n[\t\r ]*', str(self))


class Title(Payload):
    """
    Support for Mesquite's mechanism to link blocks.

    https://phylo.bio.ku.edu/slides/lab9-MesquiteAncStatesBisse/09-Mesquite.html#Create_a_tree_file
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.title = next(iter_words_and_punctuation(self._tokens, nexus=nexus)).upper()
        assert isinstance(self.title, str)


class Id(Payload):
    """
    Support for Mesquite's mechanism to link blocks.

    https://phylo.bio.ku.edu/slides/lab9-MesquiteAncStatesBisse/09-Mesquite.html#Create_a_tree_file
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.id = next(iter_words_and_punctuation(self._tokens, nexus=nexus))
        assert isinstance(self.id, str)


class Link(Payload):
    """
    Support for Mesquite's mechanism to link blocks.

    https://phylo.bio.ku.edu/slides/lab9-MesquiteAncStatesBisse/09-Mesquite.html#Create_a_tree_file
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        words = iter_words_and_punctuation(self._tokens, nexus=nexus)
        self.block = next(words).upper()
        self.title = word_after_equals(words).upper()


class Block(tuple):
    """
    A Block is a list of commands, starting with a BEGIN command and ending with END.

    .. code-block:: python

        >>> print(Block.from_commands([]))
        BEGIN BLOCK;
        END;
        >>> print(Block.from_commands([('CMD', 'A=1')]))
        BEGIN BLOCK;
            CMD A=1;
        END;
    """
    # Custom `Payload` subclasses can be registered for command names:
    __commands__ = {}

    def __new__(cls, nexus, cmds):
        return super().__new__(cls, tuple(cmds))

    def __init__(self, nexus, cmds):
        self.nexus = nexus

    def __str__(self):
        return ''.join(str(cmd) for cmd in self)

    @functools.cached_property
    def payload_map(self):
        res = {cls.__name__.upper(): cls for cls in self.__commands__}
        res.update(LINK=Link, TITLE=Title, ID=Id)
        return res

    @property
    def id(self):
        if self.ID:
            return self.ID.id

    @property
    def title(self):
        if self.TITLE:
            return self.TITLE.title

    @property
    def links(self):
        return {link.block: link.title for link in self.commands['LINK']}

    @property
    def linked_blocks(self):
        res = {}
        for name, title in self.links.items():
            for block in self.nexus.blocks.get(name, []):
                if block.title == title:
                    res[name] = block
        return res

    def __getattribute__(self, name):
        if name.isupper():
            try:
                return self.commands[name][0]
            except IndexError:
                return None
        return super().__getattribute__(name)

    @functools.cached_property
    def name(self):
        return get_name(self[0].iter_payload_tokens())

    @functools.cached_property
    def commands(self):
        res = collections.defaultdict(list)
        for cmd in self:
            if not (cmd.is_beginblock or cmd.is_endblock):
                cls = self.payload_map.get(cmd.name, Payload)
                res[cmd.name].append(cls(tuple(cmd.iter_payload_tokens()), nexus=self.nexus))
        return res

    def validate(self, log=None):
        ncmds = sum(len(cmds) for cmds in self.commands.values())
        if log:
            log.debug('{} block with {} commands'.format(self.name, ncmds))
        return True

    @classmethod
    def from_commands(cls,
                      commands: typing.Iterable[
                          typing.Union[str, typing.Tuple[str, str], typing.Tuple[str, str, str]]],
                      nexus=None,
                      name=None,
                      comment=None,
                      TITLE=None,
                      LINK=None,
                      ID=None) -> 'Block':
        """
        Generic factory method for blocks.

        This method will create a block with the uppercase name of the ``cls`` as name (or the
        explicitly passed block ``name``). The (name str, payload str) tuples from ``commands``
        are simply passed to :meth:`commonnexus.command.Command.from_name_and_payload` to assemble
        the commands in the block.

        This method should be used to create custom, non-public NEXUS blocks, while for public
        blocks the ``from_data`` method of the class implementing the block should be preferred,
        because the latter will make sure that consistent, valid block data is written.

        :param commands:
        :param nexus:
        :param name:
        :return:

        .. code-block:: python

            >>> from commonnexus import Nexus, Block
            >>> nex = Nexus()
            >>> nex.append_block(Block.from_commands([('mycommand', 'with data')], name='myblock'))
            >>> print(nex)
            #NEXUS
            BEGIN myblock;
                mycommand with data;
            END;
            >>> str(nex.MYBLOCK.MYCOMMAND)
            'with data'
        """
        cmds = [
            Command.from_name_and_payload('BEGIN', name or cls.__name__.upper(), comment=comment)]
        if TITLE:
            cmds.append(
                Command.from_name_and_payload(
                    'TITLE', Word(TITLE).as_nexus_string(), in_block=True))
        if ID:
            cmds.append(
                Command.from_name_and_payload('ID', Word(ID).as_nexus_string(), in_block=True))
        if LINK:
            if isinstance(LINK, str):
                block, _, title = LINK.partition('=')
                LINK = (block.strip(), title.strip())
            else:
                assert isinstance(LINK, tuple) and len(LINK) == 2
            cmds.append(Command.from_name_and_payload('LINK', '{} = {}'.format(
                Word(LINK[0]).as_nexus_string(), Word(LINK[1]).as_nexus_string()), in_block=True))
        for cmdspec in commands:
            payload, comment = None, None
            if isinstance(cmdspec, str):
                cname = cmdspec
            elif len(cmdspec) == 2:
                cname, payload = cmdspec
            elif len(cmdspec) == 3:
                cname, payload, comment = cmdspec
            else:
                raise ValueError(cmdspec)  # pragma: no cover
            cmds.append(
                Command.from_name_and_payload(cname, payload, comment=comment, in_block=True))
        cmds.append(Command.from_name_and_payload('END'))
        return cls(nexus, tuple(cmds))

    @classmethod
    def from_data(cls, *args, **kw) -> 'Block':
        """
        Block implementations must overwrite this method to implement "meaningful" NEXUS writing
        functionality.
        """
        raise NotImplementedError()  # pragma: no cover



"""
Combine data from multiple NEXUS files and put it in a new one.

The following blocks can be handled:

 - TAXA: Taxa are identified across NEXUS files based on label (not number).
 - CHARACTERS/DATA: Characters are aggregated across NEXUS files (with character labels prefixed,
   for disambiguation).
 - TREES: Trees are (translated and) aggregated across NEXUS files.

"""
import collections

from commonnexus import Nexus
from commonnexus.blocks import Taxa, Trees, Characters

SUPPORTED_BLOCKS = {'TAXA', 'CHARACTERS', 'DATA', 'TREES'}


def combine(*nexus: Nexus, **kw) -> Nexus:
    """
    :param nexus: `Nexus` objects to be combined.
    :return: A new `Nexus` object with the combined data.
    """
    # Make sure we are only dealing with blocks (and datatypes) that we know how to handle.
    for nex in nexus:
        if not kw.get('drop_unsupported', False):
            assert set(nex.blocks).issubset(SUPPORTED_BLOCKS), \
                "Only {} blocks are supported.".format(SUPPORTED_BLOCKS)
        for block in SUPPORTED_BLOCKS:
            assert len(nex.blocks.get(block, [])) <= 1

    # Determine the superset of taxa.
    taxa = []
    for i, nex in enumerate(nexus, start=1):
        for taxon in (nex.taxa or []):
            if taxon not in taxa:
                taxa.append(taxon)

    # Create a super-matrix, with all taxa and all characters.
    matrices, datatypes = [], set()
    charlabels = collections.defaultdict(list)
    for i, nex in enumerate(nexus, start=1):
        if nex.characters:
            datatypes.add(nex.characters.FORMAT.datatype if nex.characters.FORMAT else 'STANDARD')
            matrices.append(nex.characters.get_matrix())
            for chars in matrices[-1].values():
                for charlabel in chars:
                    charlabels[i] = charlabel
                break
    if len(datatypes) > 1:  # pragma: no cover
        raise ValueError('Only CHARACTER or DATA blocks of the same datatype can be combined!')
    matrix = collections.OrderedDict()
    if matrices:
        for taxon in taxa:
            row = collections.OrderedDict()
            for i, m in enumerate(matrices, start=1):
                if taxon in m:
                    for charlabel, val in m[taxon].items():
                        row['{}.{}'.format(i, charlabel)] = val
                else:
                    for charlabel in charlabels[i]:
                        row['{}.{}'.format(i, charlabel)] = None
            matrix[taxon] = row

    # Add all translated trees.
    trees = []
    for i, nex in enumerate(nexus, start=1):
        if nex.TREES:
            for tree in nex.TREES.trees:
                nwk = nex.TREES.translate(tree) if nex.TREES.TRANSLATE else tree.newick
                trees.append(('{}.{}'.format(i, tree.name), nwk, tree.rooted))

    nex = Nexus()
    if taxa:
        nex.append_block(Taxa.from_data(taxa))
    if matrix:
        nex.append_block(Characters.from_data(matrix))
    if trees:
        nex.append_block(Trees.from_data(*trees))

    return nex



import sys
import logging
import pathlib
import argparse
import importlib
import contextlib

import commonnexus
import commonnexus.commands


def get_log(name, level=logging.INFO) -> logging.Logger:
    logging.basicConfig(level=level)
    log = logging.getLogger(name)
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter('%(levelname)s:%(name)s:%(message)s'))
    log.addHandler(handler)
    log.setLevel(level)
    log.propagate = False
    return log


class Logging(object):
    """
    A context manager to execute a block of code at a specific logging level.
    """

    def __init__(self, logger, level=logging.DEBUG):  # pragma: no cover
        self.level = level
        self.logger = logger
        self.prev_level = self.logger.getEffectiveLevel()
        root = logging.getLogger()
        self.root_logger_level = root.getEffectiveLevel()
        self.root_handler_level = root.handlers[0].level if root.handlers else logging.WARNING

    def __enter__(self):  # pragma: no cover
        self.logger.setLevel(self.level)
        if self.logger.handlers:
            self.logger.handlers[0].setLevel(self.level)
        root = logging.getLogger()
        root.setLevel(self.level)
        if root.handlers:
            root.handlers[0].setLevel(self.level)
        return self.logger

    def __exit__(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        self.logger.setLevel(self.prev_level)
        root = logging.getLogger()
        root.setLevel(self.root_logger_level)
        if root.handlers:
            root.handlers[0].setLevel(self.root_handler_level)


class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass


def main(args=None, catch_all=False, parsed_args=None, log=None):
    from commonnexus.cli_util import ParserError

    parser = argparse.ArgumentParser(
        prog=commonnexus.__name__,
        description="{} {} is a set of commands to manipulate of files in the NEXUS "
                    "file format.".format(commonnexus.__name__, commonnexus.__version__),
        epilog='See https://github.com/dlce-eva/commonnexus for details.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # Make logging configurable:
    parser.add_argument('--log', default=get_log(commonnexus.__name__), help=argparse.SUPPRESS)
    parser.add_argument(
        '--log-level',
        default=logging.INFO,
        help='log level [ERROR {}|WARNING {}|INFO {}|DEBUG {}]'.format(
            logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG),
        type=lambda x: getattr(logging, x))
    # Register subcommands:
    subparsers = parser.add_subparsers(
        title="available commands",
        dest="_command",
        description='Run "{} COMAMND -h" to get help for a specific command.'.format(
            commonnexus.__name__),
        metavar="COMMAND")
    for p in sorted(
            pathlib.Path(__file__).parent.joinpath('commands').glob('*.py'),
            key=lambda pp: pp.stem):
        if p.stem == '__init__':
            continue
        mod = importlib.import_module('.{}'.format(p.stem), commonnexus.commands.__name__)
        help = mod.help() if hasattr(mod, 'help') else mod.__doc__
        subparser = subparsers.add_parser(
            p.stem, help=help.strip().splitlines()[0], description=help, formatter_class=Formatter)
        if hasattr(mod, 'register'):
            mod.register(subparser)
        subparser.set_defaults(main=mod.run)

    args = parsed_args or parser.parse_args(args=args)

    if not hasattr(args, "main"):
        parser.print_help()
        return 1

    with contextlib.ExitStack() as stack:
        if not log:  # pragma: no cover
            stack.enter_context(Logging(args.log, level=args.log_level))
        else:
            args.log = log
        try:
            return args.main(args) or 0
        except KeyboardInterrupt:  # pragma: no cover
            return 0
        except ParserError as e:  # pragma: no cover
            print(str(e))
            return main([args._command, '-h'])
        except Exception as e:  # pragma: no cover
            if catch_all:
                print(e)
                return 1
            raise


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main() or 0)



"""
Manipulate a TREES block in a NEXUS file.

Note: Only one option can be chosen at a time.
"""
import enum
import random

from commonnexus.blocks import Trees
from commonnexus.cli_util import (
    add_nexus, add_flag, list_of_ranges, validate_operations, add_rename,
)


class Ops(enum.Enum):
    drop = 1
    sample = 2
    random = 3
    strip_comments = 4
    rename = 5


def register(parser):
    add_nexus(parser)
    parser.add_argument(
        "--drop",
        default=[],
        type=list_of_ranges,
        help="Remove the trees specified as comma-separated ranges of 1-based indices, "
             "e.g. '1', '1-5', '1,20-30'")
    parser.add_argument(
        "--sample",
        type=int,
        metavar='N',
        default=0,
        help="Resample the trees every Nth tree")
    parser.add_argument(
        "--random",
        type=int,
        metavar='N',
        default=0,
        help="Randomly sample N trees from the treefile")
    parser.add_argument(
        '--random-seed',
        help="Set random seed (to a number) to allow for reproducible random sampling.",
        type=lambda s: random.seed(int(s)))
    parser.add_argument(
        "--strip-comments",
        action="store_true",
        default=False,
        help="Remove comments from the trees")
    add_rename(parser, 'tree')
    add_flag(parser, 'describe', 'list 1-based index, names and rooting of trees')


def run(args):
    validate_operations(Ops, args)
    if args.describe:
        args.log.info('Trees\nindex\tname\trooted')
        for i, tree in enumerate(args.nexus.TREES.trees, start=1):
            print('{}\t{}\t{}'.format(i, tree.name, tree.rooted))
        return

    trees = [cmd for cmd in args.nexus.TREES if cmd.name == 'TREE']
    if args.drop:
        for i, tree in enumerate(trees, start=1):
            if i in args.drop:
                args.nexus.remove(tree)
    elif args.sample:
        for i, tree in enumerate(trees, start=1):
            if i % args.sample != 0:
                args.nexus.remove(tree)
    elif args.random:
        sampled = random.sample(trees, args.random)
        for tree in trees:
            if tree not in sampled:
                args.nexus.remove(tree)
    elif args.strip_comments:
        trees = []
        for tree in args.nexus.TREES.trees:
            trees.append((tree.name, tree.newick.strip_comments(), tree.rooted))
        labels = args.nexus.TREES.TRANSLATE.mapping if args.nexus.TREES.TRANSLATE else {}
        args.nexus.replace_block(args.nexus.TREES, Trees.from_data(*trees, **labels))
    elif args.rename:
        trees = []
        for number, tree in enumerate(args.nexus.TREES.trees, start=1):
            tname = tree.name
            if callable(args.rename):
                tname = args.rename(tree.name)
            elif args.rename[0] == str(number) or args.rename[0] == tree.name:
                tname = args.rename[1]
            trees.append((tname, tree.newick, tree.rooted))
        labels = args.nexus.TREES.TRANSLATE.mapping if args.nexus.TREES.TRANSLATE else {}
        args.nexus.replace_block(args.nexus.TREES, Trees.from_data(*trees, **labels))

    print(args.nexus)



"""
Get help on subcommands.
"""


def register(parser):
    parser.add_argument('command', nargs='?', default=None)


def run(args):
    from commonnexus.__main__ import main

    if args.command:
        return main([args.command, '-h'])
    print('Run "commonnexus COMMAND -h" to get help on subcommand COMMAND')



from .base import Block


class Unaligned(Block):
    """
    .. warning::

        `commonnexus` doesn't provide any functionality - other than parsing as generic commands -
        for ``UNALIGNED`` blocks yet.

    The UNALIGNED block includes data that are not aligned. Its primary intent is to house
    unaligned DNA, RNA, NUCLEOTIDE, and PROTEIN sequence data. Taxa are usually not defined in an
    UNALIGNED block; if not, this block must be preceded by a block that defines taxon labels and
    ordering (e.g., TAXA).
    Syntax of the UNALIGNED block is as follows:

    .. rst-class:: nexus

        | BEGIN UNALIGNED;
        |   [DIMENSIONS NEWTAXA NTAX=number-of-taxa;]
        |   [FORMAT
        |     [DATATYPE = { STANDARD | DNA | RNA | NUCLEOTIDE | PROTEIN}]
        |     [RESPECTCASE]
        |     [MISSING=symbol]
        |     [SYMBOLS="symbol [symbol...]"]
        |     [EQUATE = "symbol=entry [symbol=entry...]"]
        |     [[NO]LABELS]
        |   ;]
        |   [TAXLABELS taxon-name [taxon-name...];]
        |   MATRIX data-matrix;
        | END;

    Commands must appear in the order listed. Only one of each command is allowed per block.
    The DIMENSIONS command should only be included if new taxa are being defined in this block,
    which is discouraged (see discussion under DATA block). The format for the DIMENSIONS command
    is as in the CHARACTERS block, except NCHAR is not allowed.
    Subcommands of the FORMAT command are described in the CHARACTERS block. The TAXLABELS command
    serves to define taxa and is only allowed if the NEWTAXA token is included in the DIMENSIONS
    statement. It follows the same form as described in the TAXA block.
    Here is an example of an UNALIGNED block

    .. code-block::

        BEGIN UNALIGNED;
            FORMAT DATATYPE=DNA;
            MATRIX
                taxon_1 ACTAGGACTAGATCAAGTT,
                taxon_2 ACCAGGACTAGCGGATCAAG,
                taxon_3 ACCAGGACTAGATCAAG,
                taxon_4 AGCCAGGACTAGTTC,
                taxon_5 ATCAGGACTAGATCAAGTTC;
        END;

    A comma must be placed at the end of each sequence (except the last, which requires a
    semicolon). Each sequence can occupy more than one line.
    """



"""
Split a Mesquite multi-block NEXUS into individual NEXUS files per CHARACTERS/TREES block.

Output files will be named according to the pattern "<stem>_<BLOCK.TITLE>.{nex|trees}".
So a CHARACTERS block with TITLE 'x.y' will end up in 'mesquite_x.y.nex'.
"""
import pathlib

from commonnexus.cli_util import add_nexus
from commonnexus import Nexus


def register(parser):
    add_nexus(parser)
    parser.add_argument(
        '--stem',
        help='Stem of the filenames for the individual nexus files',
        default='mesquite')
    parser.add_argument(
        '--outdir',
        help="(Existing) directory to write the output to.",
        default='.')


def run(args):
    for block, suffix in dict(TREES='trees', CHARACTERS='nex').items():
        for i, block in enumerate(args.nexus.blocks[block], start=1):
            blocks = [block]
            if 'TAXA' in block.linked_blocks:
                blocks.insert(0, block.linked_blocks['TAXA'])
            nex = Nexus.from_blocks(*blocks)
            p = pathlib.Path(args.outdir) / '{}_{}.{}'.format(args.stem, block.title or i, suffix)
            nex.to_file(p)
            args.log.info('CHARACTERS block written to {}'.format(p))



"""
Normalise a NEXUS file.

Normalisation includes

 - converting CHARACTERS/DATA matrices to non-transposed, non-interleaved representation with
   taxon labels (and resolved EQUATEs), extracting taxon labels into a TAXA block;
 - converting a DISTANCES matrix to non-interleaved matrices with diagonal and both triangles
   and taxon labels;
 - translating all TREEs in a TREES block (such that the TRANSLATE command becomes superfluous).

In addition, after normalisation, the following assumptions hold:

- All commands start on a new line.
- All command names (**not** block names) are in uppercase with no "in-name-comment",
  like "MA[c]TRiX"
- The ";" terminating MATRIX commands is on a separate line, allowing more simplistic parsing
  of matrix rows.
"""
import typing
import collections

from commonnexus import Nexus
from commonnexus.blocks.characters import Data
from commonnexus.blocks import Taxa, Distances, Characters, Trees


def normalise(nexus: Nexus,
              data_to_characters: bool = False,
              strip_comments: bool = False,
              remove_taxa: typing.Optional[typing.Container[str]] = None) -> Nexus:
    """
    :param nexus: A `Nexus` object to be normalised in-place.
    :param data_to_characters: Flag signaling whether DATA blocks should be converted to CHARACTER \
    blocks.
    :param strip_comments: Flag signaling whether to remove all non-command comments.
    :param remove_taxa: Container of taxon labels specifying taxa to remove from relevant blocks.
    :return: The modified `Nexus` object.

    .. code-block:: python

        >>> from commonnexus import Nexus
        >>> from commonnexus.tools import normalise
        >>> print(normalise(Nexus('''#NEXUS
        ... BEGIN CHARACTERS;
        ... DIMENSIONS NCHAR=3;
        ... FORMAT DATATYPE=STANDARD MISSING=x GAP=- SYMBOLS="01" INTERLEAVE;
        ... MATRIX
        ...     t1 10
        ...     t2    01
        ...     t3    00
        ...     t1 0
        ...     t2 0
        ...     t3 1;
        ... END;
        ... BEGIN DISTANCES;
        ... DIMENSIONS NTAX=3;
        ... FORMAT NODIAGONAL MISSING=?;
        ... MATRIX
        ...     t1
        ...     t2    1.0
        ...     t3    2.0 3.0;
        ... END;
        ... BEGIN TREES;
        ... TRANSLATE a t1, b t2, c t3;
        ... TREE 1 = (a,b,c);
        ... END;''')))
        #NEXUS
        BEGIN TAXA;
        DIMENSIONS NTAX=3;
        TAXLABELS t1 t2 t3;
        END;
        BEGIN CHARACTERS;
        DIMENSIONS NCHAR=3;
        FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
        MATRIX
        t1 100
        t2 010
        t3 001
        ;
        END;
        BEGIN DISTANCES;
        DIMENSIONS NTAX=3;
        FORMAT TRIANGLE=BOTH MISSING=?;
        MATRIX
        t1 0 1.0 2.0
        t2 1.0 0 3.0
        t3 2.0 3.0 0
        ;
        END;
        BEGIN TREES;
        TREE 1 = (t1,t2,t3);
        END;
    """
    remove_taxa = remove_taxa or []

    if strip_comments:
        nexus = Nexus([cmd.without_comments() for cmd in nexus], config=nexus.cfg)
    nexus = Nexus([cmd.with_normalised_whitespace() for cmd in nexus], config=nexus.cfg)

    taxlabels = None
    if nexus.characters:
        matrix = nexus.characters.get_matrix()
        taxlabels = list(matrix.keys())
        matrix = collections.OrderedDict((k, v) for k, v in matrix.items() if k not in remove_taxa)
        characters = nexus.DATA or nexus.CHARACTERS
        cls = Data if characters.name == 'DATA' and not data_to_characters else Characters
        nexus.replace_block(
            characters,
            cls.from_data(matrix, statelabels=nexus.characters.get_charstatelabels()[1]))

    if nexus.DISTANCES:
        matrix = nexus.DISTANCES.get_matrix()
        if taxlabels:
            assert set(matrix.keys()).issubset(taxlabels)
        else:
            taxlabels = list(matrix.keys())
        matrix = collections.OrderedDict(
            (k, collections.OrderedDict((kk, vv) for kk, vv in v.items() if kk not in remove_taxa))
            for k, v in matrix.items() if k not in remove_taxa)
        nexus.replace_block(nexus.DISTANCES, Distances.from_data(matrix))

    if nexus.TREES:
        trees = []
        for tree in nexus.TREES.trees:
            nwk = nexus.TREES.translate(tree) if nexus.TREES.TRANSLATE else tree.newick
            if remove_taxa:
                nwk.prune_by_names(remove_taxa)
            trees.append((tree.name, nwk, tree.rooted))
        nexus.replace_block(nexus.TREES, Trees.from_data(*trees))

    if taxlabels:
        taxa = Taxa.from_data([t for t in taxlabels if t not in remove_taxa])
        if nexus.TAXA:
            assert nexus.TAXA.DIMENSIONS.ntax == len(taxlabels)
            assert set(nexus.TAXA.TAXLABELS.labels.values()) == set(taxlabels)
            nexus.replace_block(nexus.TAXA, taxa)
        else:
            nexus.prepend_block(taxa)
    return nexus
